import os
from pathlib import Path
import numpy as np


def check_rm(path):
    """
    Check and delete file

    :param path: path string of target file
    """
    if os.path.isfile(path):
        print(f"removing {path}")
        os.remove(path)


def file_exists(path):
    if os.path.isfile(path):
        return True
    else:
        raise FileNotFoundError(f"{path} not exist")


def is_empty(path):
    """
    Check if file is empty. If the file does not exist, then create one.

    :param path: path string of target file
    :returns: boolean operator
    """
    if os.path.isfile(path):
        return os.stat(path).st_size == 0
    else:
        Path(path).touch()


def set_sns_barwidth(ax, width):
    width_list = [width] * len(ax.patches)

    for bar, newwidth in zip(ax.patches, width_list):
        # get current position
        x = bar.get_x()
        width = bar.get_width()
        centre = x + width / 2.0

        # update width
        bar.set_x(centre - newwidth / 2.0)
        bar.set_width(newwidth)


def mean_w_na(data, na_policy):
    if na_policy == "zero":
        return np.mean(np.nan_to_num(data))
    elif na_policy == "ignore":
        return np.nanmean(data)


def remove_outliers(data, m=5):
    """
    remove extreme value (outliers) based on median absolute deviation (MAD) measurement around the median
    :param data: numpy array
    :param m: tolerance, the larger `m` remove less outliers
    :returns: numpy array with outlier removed
    """
    distance = np.abs(data - np.nanmedian(data))
    mdev = np.nanmedian(distance)
    s = distance / (mdev if mdev else 1.0)
    np.putmask(data, s > m, np.nan)
    return data


def powlaw(x, a, b):
    return a * np.power(x, b)


def esd_pow(n=8, esd_max=1900):
    """
    ecosystem size structure builder following power law, size=k(n^a).
    build from scipy.curve_fit based on exisiting 8P8Z and 32P32Z size structure.
    Principal is more size classes (`n`) should allocate more at smaller organisms
    instead of interpolation or equal increment.

    :param n: number of size classes
    :param esd_max: maximum individual size (um)
    """

    a = 8.5 + 0.25 * (n / 8)
    k = esd_max / n**a
    x = powlaw(np.linspace(1, n, n), k, a)
    return x


def esd_ward(N=2, esd_min=0.6, k=1.3, esd_max=1900):
    """
    Ward et al. (2018) method
    :param N: the least size classes in each loop
    :param k: size increment
    :param esd_max: maximum, in micron
    """

    # build a smallest size group
    l = []
    for i in range(N):
        l.append(esd_min + i * k)
        esd = np.array(l)

    # growth by a exponent of 10
    while esd.max() < esd_max:
        esd = np.append(esd, esd[-N:] * 10)
    return esd


def distance(x, y):
    """
    euclidean distance for 2D array

    x/y: delta_value array in 2D
    """
    return np.sqrt(np.square(x) + np.square(y))



import pathlib

import numpy as np
import pandas as pd
import cartopy.crs as ccrs
from cartopy.feature import LAND
from matplotlib.colors import ListedColormap
import matplotlib.pyplot as plt
from matplotlib.projections import PolarAxes
import mpl_toolkits.axisartist.floating_axes as fa
import mpl_toolkits.axisartist.grid_finder as gf

from .data import efficient_log
from .grid import GENIE_lat, GENIE_lon, GENIE_depth

def plot_genie(
    ax,
    data,
    log=False,
    grid_line=False,
    continent_outline=True,
    contour_layer=False,
    *args,
    **kwargs,
):
    """
    plot map for 2D GENIE time-slice data

    :param data: 2D numpy array
    :param ax: matplotlib axes object with cartopy projection. e.g., projection=ccrs.EckertIV()
    :param vmin: colorbar limit
    :param log: whether take log for data
    :type log: boolean
    :param grid_line: whthere plot grid line
    :type grid_line: boolean
    :param continent_outline: logic whethere plot continent

    :returns: a mappable plot object
    """

    data_crs = ccrs.PlateCarree()  # should not change

    # -------------------Box line------------------------
    ax.spines["geo"].set_edgecolor("black")
    ax.spines["geo"].set_linewidth(1)

    # -------------------Facecolor-----------------------
    ax.patch.set_color("silver")

    # -------------------Plot-----------------------
    if log:
        data = efficient_log(data)

    lon_edge = GENIE_lon(edge=True)
    lat_edge = GENIE_lat(edge=True)

    # cartopy transform seems to help reassign the GENIE longitude to normal
    p = ax.pcolormesh(
        lon_edge,
        lat_edge,
        data,
        transform=data_crs,
        shading="flat",
        *args,
        **kwargs,
    )

    if contour_layer:
        lat = GENIE_lat(edge=False)
        lon = GENIE_lon(edge=False)
        cs = ax.contour(lon, lat, data, transform=ccrs.PlateCarree(), **kwargs)
        # label every three levels
        ax.clabel(cs, cs.levels[::3], colors=["black"], fontsize=8, inline=False)

    # -------------------Grid lines-----------------------
    if grid_line:
        ax.gridlines(
            crs=data_crs,
            draw_labels=False,
            linewidth=0.5,
            color="gray",
            alpha=0.5,
            linestyle="-",
        )

    # -------------------Continent lines-----------------------
    if continent_outline:
        outline_color = "black"
        outline_width = 1
        mask_array = np.where(~np.isnan(data), 1, 0)

        Nlat_edge = len(lat_edge)
        Nlon_edge = len(lon_edge)
        # dimension of array
        Nlat_array = Nlat_edge - 1
        Nlon_array = Nlon_edge - 1
        # index of array
        Nlat_index = Nlat_array - 1
        Nlon_index = Nlon_array - 1

        # i stands for latitude index, j stands for longitude index
        # e.g., mask_array[0, ] is Antarctic ice cap
        # i,j are for mask array and will be converted to lon/lat edge lines
        for i in range(Nlat_array):
            for j in range(Nlon_array):
                # compare with the right grid, and plot vertical line if different
                if j < Nlon_index and mask_array[i, j] != mask_array[i, j + 1]:
                    ax.vlines(
                        lon_edge[j + 1],
                        lat_edge[i],
                        lat_edge[i + 1],
                        color=outline_color,
                        linewidth=outline_width,
                        transform=data_crs,
                    )

                # connect the circular longitude axis
                if j == Nlon_index and mask_array[i, j] != mask_array[i, 0]:
                    ax.vlines(
                        lon_edge[j + 1],
                        lat_edge[i],
                        lat_edge[i + 1],
                        color=outline_color,
                        linewidth=outline_width,
                        transform=data_crs,
                    )

                # compare with the above grid, and plot horizontal line if different
                if i < Nlat_index and mask_array[i, j] != mask_array[i + 1, j]:
                    ax.hlines(
                        lat_edge[i + 1],
                        lon_edge[j],
                        lon_edge[j + 1],
                        colors=outline_color,
                        linewidth=outline_width,
                        transform=data_crs,
                    )

    return p


def scatter_map(
    df: pd.DataFrame,
    var,
    ax,
    x="Longitude",
    y="Latitude",
    add_layer=True,
    log=False,
    *args,
    **kwargs,
):
    """plot map based on dataframe with latitude/longitude
    using cartopy as engine

    :param df: pandas dataframe
    :param var: variable (column) in dataframe to plot
    :param x: coordinate attribute, default "Longitude"
    :param y: coordinate attribute, default "Latitude"
    :param add_layer: whether to add basic costal lines, and land feature

    :returns: a map
    """

    if add_layer:
        ax.set_global()
        ax.coastlines()
        ax.add_feature(
            LAND, zorder=0, facecolor="#B1B2B4", edgecolor="white"
        )  # zorder is drawing sequence

    if "Latitude" not in df.columns or "Longitude" not in df.columns:
        raise ValueError("Input data lack Latitude/Longitude column")

    if log:
        df[var] = efficient_log(df[var])

    p = ax.scatter(
        x=df[x],
        y=df[y],
        c=df[var],
        linewidths=0.5,
        edgecolors="black",
        transform=ccrs.PlateCarree(),
        *args,
        **kwargs,
    )

    return p


def genie_cmap(cmap_name, N=256, reverse=False):
    """
    Get a self-defined colormap

    :param cmap_name: Zissou1, FantasticFox, Rushmore, Darjeeling, ODV
    :type cmap_name: str

    :returns: colormap
    """
    file_name = f"data/{cmap_name}.txt"
    file_path = pathlib.Path(__file__).parent.parent / file_name

    colors = pd.read_csv(file_path, header=None).values.tolist()
    colors = [colors[i][0] for i in range(len(colors))]

    c = ListedColormap(colors, N=N)

    if reverse:
        return c.reversed()
    return c


def cbar_wrapper(plotting_func):
    """a decorator to add color bar

    :param plotting_func: function returning a mappable object
    :returns: plotting function with colorbar
    """

    def wrappered_func(*args, **kwargs):
        p = plotting_func(*args, **kwargs)
        cbar = plt.colorbar(p, fraction=0.05, pad=0.04, orientation="vertical")
        cbar.ax.tick_params(color="k", direction="in")
        cbar.minorticks_on()

    return wrappered_func


class GeniePlottable(object):
    def plot_line(self, dim, *args, **kwargs):
        """plot 1D data, e.g., zonal_average"""

        if self.dim() > 1:
            raise ValueError("Data should be of 1 Dimension")
        if dim == "lon":
            x = GENIE_lat()
        if dim == "lat":
            x = GENIE_lon()

        fig = plt.figure(figsize=(6, 4))
        ax = fig.add_subplot(111)
        ax.minorticks_on()

        p = ax.plot(x, self.array, "k", *args, **kwargs)

        return p

    def plot_map(self, ax=None, polar=False, cbar=True, *args, **kwargs):

        """plot lat-lon 2D array"""

        plt.rcParams["font.family"] = "sans-serif"

        if polar:
            fig = plt.figure(figsize=[10, 5])
            ax_arctic = plt.subplot(121, projection=ccrs.Orthographic(0, 90))
            ax_antarctic = plt.subplot(122, projection=ccrs.Orthographic(180, -90))
            ax_arctic.gridlines(draw_labels=True)
            ax_antarctic.gridlines(draw_labels=True)
            plot_genie(ax=ax_arctic, data=self.array, *args, **kwargs)
            plot_genie(ax=ax_antarctic, data=self.array, *args, **kwargs)

            return fig

        if not ax:
            fig = plt.figure(dpi=75)
            ax = fig.add_subplot(111, projection=ccrs.EckertIV())

        if np.any(self.array < 0) and np.any(self.array > 0):
            # x = max(abs(self.nanmin()), self.nanmax())
            # vmax = x/2
            # vmin = vmax * -1
            p = plot_genie(ax=ax, data=self.array, cmap="RdBu_r", *args, **kwargs)
        else:
            p = plot_genie(ax=ax, data=self.array, *args, **kwargs)

        if cbar:
            cax = fig.add_axes([0.15, 0.1, 0.73, 0.07])  # xmin, ymin, dx, dy
            cbar = fig.colorbar(p, cax=cax, orientation="horizontal", pad=0.04)
            cbar.minorticks_on()
            if hasattr(self, "unit"):
                cbar.set_label(self.unit, size=12)

        return p

    def cross_section(self, ax=None, cmap="YlGnBu_r", *args, **kwargs):
        """plot 3D array, tracers and stream function"""

        data = self.array

        # visualise
        if not ax:
            fig, ax = plt.subplots(figsize=(8, 4))

        lat_edge = GENIE_lat(edge=True)
        z_edge = GENIE_depth(edge=True) / 1000
        lat = GENIE_lat(edge=False)
        z = GENIE_depth(edge=False) / 1000

        # pcolormesh
        p = ax.pcolormesh(lat_edge, z_edge, data, cmap=cmap, *args, **kwargs)

        # contourf
        # contourf_max = np.max(moc)//10 * 10
        # controuf_N = int(contourf_max*2/10 + 1)
        # contourf_level = np.linspace(contourf_max*-1, contourf_max, controuf_N)
        # cs = ax.contourf(lat, z, data, corner_mask=False, levels=contourf_level, cmap=contourf_cmap)

        # contour
        cs = ax.contour(
            lat, z, data, linewidths=0.6, colors="black", linestyles="solid"
        )
        ax.clabel(cs, cs.levels[::1], colors=["black"], fontsize=8.5, inline=False)

        # colorbar, ticks & labels
        # divider = make_axes_locatable(ax)
        # cax = divider.append_axes('right', size='5%', pad=0.05)
        cbar = plt.colorbar(p, fraction=0.05, pad=0.04, orientation="vertical")
        cbar.ax.tick_params(color="k", direction="in")
        # cbar.set_label("Stream function (Sv)")

        plt.rcParams["font.family"] = "Arial"
        ax.patch.set_color("#999DA0")
        ax.set_ylim(ax.get_ylim()[::-1])
        ax.set_xlabel(r"Latitude ($\degree$N)", fontsize=13)
        ax.set_ylabel("Depth (km)", fontsize=12)
        ax.minorticks_on()
        ax.tick_params("both", length=4, width=1, which="major")

        return p


class TaylorDiagram(object):
    """
    Taylor diagram.
    Plot model standard deviation and correlation to reference data in a single-quadrant polar plot,
    with r=stddev and theta=arccos(correlation).

    modified from Yannick Copin, https://gist.github.com/ycopin/3342888
    reference: https://matplotlib.org/stable/gallery/axisartist/demo_floating_axes.html
    """

    def __init__(
        self,
        fig=None,
        figscale=1,
        subplot=111,
        xmax=None,
        tmax=np.pi / 2,
        ylabel="Standard Deviation",
        rotation=None,
    ):

        """
        Set up Taylor diagram axes, i.e. single quadrant polar
        plot, using `mpl_toolkits.axisartist.floating_axes`.

        Parameters:

        * fig: input Figure or None
        * subplot: subplot definition
        * xmax: the length of radius, xmax can be 1.5* reference std
        """

        # --------------- tickers --------------------------
        # Correlation labels (if half round)
        cor_label = np.array([0, 0.2, 0.4, 0.6, 0.8, 0.9, 0.95, 0.99, 1])

        # add the negative ticks if more than half round
        excess_theta = tmax - np.pi / 2
        if excess_theta > 0:
            cor_label = np.concatenate((-cor_label[:0:-1], cor_label))

        # convert to radian
        rad = np.arccos(cor_label)
        # tick location
        gl = gf.FixedLocator(rad)
        # tick formatting: bind radian and correlation coefficient
        tf = gf.DictFormatter(dict(zip(rad, map(str, cor_label))))

        # --------------- coordinate -----------------------
        # Standard deviation axis extent (in units of reference stddev)
        # xmin must be 0, which is the centre of round

        self.xmin = 0
        self.xmax = xmax
        self.tmax = tmax

        # ------- curvilinear coordinate definition -------
        # use built-in polar transformation (i.e., from theta and r to x and y)
        tr = PolarAxes.PolarTransform()
        ghelper = fa.GridHelperCurveLinear(
            tr,
            extremes=(0, self.tmax, self.xmin, self.xmax),
            grid_locator1=gl,
            tick_formatter1=tf,
        )

        # ------- create floating axis -------
        if fig is None:
            fig_height = 4.5 * figscale
            fig_width = fig_height * (1 + np.sin(excess_theta))
            fig = plt.figure(figsize=(fig_width, fig_height), dpi=100)

        ax = fa.FloatingSubplot(fig, subplot, grid_helper=ghelper)
        fig.add_subplot(ax)

        # Adjust axes
        # Angle axis
        ax.axis["top"].label.set_text("Correlation")
        ax.axis["top"].toggle(ticklabels=True, label=True)
        # inverse the direction
        ax.axis["top"].set_axis_direction("bottom")
        ax.axis["top"].major_ticklabels.set_axis_direction("top")
        ax.axis["top"].label.set_axis_direction("top")

        # X axis
        ax.axis["left"].set_axis_direction("bottom")

        # Y axis direction & label
        ax.axis["right"].toggle(all=True)
        ax.axis["right"].label.set_text(ylabel)
        ax.axis["right"].set_axis_direction("top")
        # ticklabel direction
        ax.axis["right"].major_ticklabels.set_axis_direction("left")

        ax.axis["bottom"].set_visible(False)

        # ------- Set instance attribute ----------
        self.fig = fig
        # Graphical axes
        self._ax = ax
        # grid line
        self._ax.grid(True, zorder=0, linestyle="--")
        # aspect ratio
        self._ax.set_aspect(1)
        # A parasite axes for further plotting data
        self.ax = ax.get_aux_axes(tr)
        # Collect sample points for latter use (e.g. legend)
        self.samplePoints = []

    def add_ref(self, refstd, reflabel="Observation", linestyle="-", color="k"):
        """add a reference point"""
        self.refstd = refstd
        # Add reference point
        # slightly higher than 0 so star can be fully seen
        l = self.ax.plot(0.01, self.refstd, "k*", ls="", ms=10)
        # xy for the point, xytext for the text (the coordinates are
        # defined in xycoords and textcoords, respectively)
        self.ax.annotate(
            reflabel,
            xy=(0.01, self.refstd),
            xycoords="data",
            xytext=(-25, -30),
            textcoords="offset points",
        )
        # add stddev contour
        t = np.linspace(0, self.tmax)
        r = np.zeros_like(t) + self.refstd
        self.ax.plot(t, r, linestyle=linestyle, color=color)
        self.samplePoints.append(l)

    def add_scatter(self, stddev, corrcoef, *args, **kwargs):
        """
        Add sample (*stddev*, *corrcoeff*) to the Taylor
        diagram. *args* and *kwargs* are directly propagated to the
        `Figure.plot` command.
        """

        l = self.ax.scatter(
            np.arccos(corrcoef), stddev, *args, **kwargs
        )  # (theta, radius)
        self.samplePoints.append(l)

        return l

    def add_contours(self, levels=5, **kwargs):
        """
        Add constant centered RMS difference contours, defined by *levels*.
        """

        rs, ts = np.meshgrid(
            np.linspace(self.xmin, self.xmax), np.linspace(0, self.tmax)
        )
        # Compute centered RMS difference
        crmse = np.sqrt(self.refstd**2 + rs**2 - 2 * self.refstd * rs * np.cos(ts))
        contours = self.ax.contour(ts, rs, crmse, levels, linestyles="--", **kwargs)
        self.ax.clabel(contours, contours.levels[::1], inline=False)

        return contours

    def add_legend(self, *args, **kwargs):
        return self.ax.legend(*args, **kwargs)

    def add_annotation(self, *args, **kwargs):
        return self.ax.annotation(*args, **kwargs)

    def savefig(self, *args, **kwargs):
        self.fig.savefig(*args, **kwargs)



from functools import reduce
from os.path import join
import string

import xarray as xr
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator
import cartopy.crs as ccrs
import numpy as np
from scipy.stats import sem
import regionmask
from netCDF4 import Dataset
from pandas import DataFrame, read_fwf
import seaborn as sns

from . import ureg, Q_
from .plot import plot_genie, GeniePlottable
from .grid import (
    GENIE_grid_area,
    reassign_GENIE,
    GENIE_grid_mask,
    GENIE_grid_vol,
    GENIE_lat,
    GENIE_lon,
    normal_lon,
)
from .data import foram_dict, foram_names, obs_stat_bytype, obs_stat_bysource
from .scores import quick_mscore, quick_rmse, quick_cos_sim, quick_corr
from .utils import file_exists, set_sns_barwidth, distance
from .chem import rm_element, molecular_weight


class GenieArray(GeniePlottable):
    def __init__(self, M=36, N=36):
        """
        Create an empty 2D array, default as 36x36
        """
        # set dimension
        self.M = M
        self.N = N
        # set data
        self.array = self._set_array()
        # set unit
        if hasattr(self.array, "units"):
            self.unit = self.array.units
        else:
            self.unit = ""

    def _set_array(self):
        "assign real data"
        return np.zeros((self.M, self.N))

    def pure_array(self):
        "get a numpy array"
        if hasattr(self.array, "values"):
            return self.array.values
        else:
            return self.array

    def uarray(self):
        "array with unit"
        unit = rm_element(self.unit)
        uarray = Q_(self.pure_array(), unit)
        return uarray

    def dim(self):
        return self.pure_array().ndim

    def flip(self, axis=0):
        return np.flip(self.array, axis=axis)

    def flatten(self):
        "flatten in row-major (C-style)"
        return self.pure_array().flatten(order="C")

    def apply(self, f):
        vfunc = np.vectorize(f)
        x = GenieArray()
        x.array = vfunc(self.pure_array())
        return x

    def reassign_array(self):
        "if self.array is xarray, then reassign the coordinate"
        x = GenieArray()
        x.array = reassign_GENIE(self.array).to_numpy()
        return x

    def _to_genie_array(self):
        """
        remove all attributes
        """
        empty = GenieArray()
        empty.array = self.array
        return empty

    def _run_method(self, method: str, *args, **kwargs):
        "an alias to run stat for GenieArray class"
        return getattr(self, method)(*args, **kwargs)

    def __add__(self, other):
        sum = GenieArray()
        if hasattr(other, "array"):
            sum.array = self.array + other.array
        else:
            sum.array = self.array + other
        return sum

    def __sub__(self, other):
        diff = GenieArray()
        if hasattr(other, "array"):
            diff.array = self.array - other.array
        else:
            diff.array = self.array - other
        return diff

    def __truediv__(self, other):
        quotient = GenieArray()
        if hasattr(other, "array"):
            quotient.array = np.divide(
                self.array,
                other.array,
                out=np.zeros_like(self.array),
                where=other.array != 0,
            )
        else:
            try:
                quotient.array = np.divide(self.array, other)
            except ValueError:
                print("Sorry, either number and GenieArray are accepted")

        return quotient

    def __mul__(self, other):
        product = GenieArray()
        if hasattr(other, "array"):
            product.array = self.array * other.array
        else:
            try:
                product.array = self.array * other
            except ValueError:
                print("Sorry, only number and GenieArray are accepted")

        return product

    def max(self, *args, **kwargs):
        return np.max(self.array, *args, **kwargs)

    def nanmax(self, *args, **kwargs):
        return np.nanmax(self.array, *args, **kwargs)

    def min(self, *args, **kwargs):
        return np.min(self.array, *args, **kwargs)

    def nanmin(self, *args, **kwargs):
        return np.nanmin(self.array, *args, **kwargs)

    def sum(self, *args, **kwargs):
        return np.sum(self.pure_array(), *args, **kwargs)

    def square(self, *arg, **kwargs):
        return np.square(self.pure_array())

    def sqrt(self, *arg, **kwargs):
        return np.sqrt(self.pure_array())

    def nansum(self, *args, **kwargs):
        return np.nansum(self.pure_array(), *args, **kwargs)

    def mean(self, *args, **kwargs):
        return np.mean(self.pure_array(), *args, **kwargs)

    def nanmean(self, *args, **kwargs):
        return np.nanmean(self.pure_array(), *args, **kwargs)

    def ptp(self, *args, **kwargs):
        return np.ptp(self.pure_array(), *args, **kwargs)

    def sd(self, *args, **kwargs):
        return np.std(self.pure_array(), *args, **kwargs)

    def nansd(self, *args, **kwargs):
        return np.nanstd(self.pure_array(), *args, **kwargs)

    def cv(self):
        "coefficient of variance, or normalized standard deviation"
        cv = self.nansd() / self.nanmean()
        return cv

    def se(self, *args, **kwargs):
        return sem(self.array, nan_policy="omit", axis=None, *args, **kwargs)

    def select_basin(self, basin):
        ocean = regionmask.defined_regions.ar6.ocean
        index = ocean.map_keys(basin)
        mask = ocean.mask(self.reassign_array())
        regional_data = self.reassign_array().where(mask == index)

        return regional_data

    def mask_basin(self, base, basin, basin_lvl):
        # mask data
        data = self.pure_array()
        mask = GENIE_grid_mask(base=base, basin=basin, basin_lvl=basin_lvl, invert=True)

        if self.dim() > 2:
            mask = np.broadcast_to(mask, (16, 36, 36))

        mask_data = np.ma.array(data, mask=mask)
        mask_data = np.ma.masked_invalid(mask_data)

        garray = GenieArray()
        garray.array = mask_data

        return garray

    def mean_along(self, dim):
        # time, zt, lat, lon
        if self.dim() == 2:
            if dim == "lat":
                ax = 0
            elif dim == "lon":
                ax = 1
        elif self.dim() == 3:
            if dim == "z":
                ax = 0
            if dim == "lat":
                ax = 1
            elif dim == "lon":
                ax = 2

        garray = GenieArray()
        garray.array = self.nanmean(axis=ax)

        return garray

    def search_grid(self, *args, **kwargs):
        return self.array.sel(*args, **kwargs, method="nearest")

    def search_range(self, lon_min=-255, lon_max=95, lat_min=0, lat_max=90):
        """
        default longitude is unassigned of cGENIE grids
        """

        if lon_min > lon_max or lat_min > lat_max:
            raise ValueError("longitude/latitude min must be less than max!")

        lon = self.array.coords["lon"]
        lat = self.array.coords["lat"]

        return self.array.loc[
            dict(
                lat=lat[(lat >= lat_min) & (lat <= lat_max)],
                lon=lon[(lon >= lon_min) & (lon <= lon_max)],
            )
        ]

    def filter(self, threshold, greater_sign=True):
        data = self.array

        if greater_sign:
            return data.where(data > threshold, drop=True)
        else:
            return data.where(data < threshold, drop=True)


class GenieModel(object):
    def __init__(self, model_path):
        self.model_path = model_path
        self._time = -1

    @property
    def time(self):
        return self._time

    @time.setter
    def time(self, value):
        if not isinstance(value, int):
            time_array = self.vars(var_name="time")
            raise ValueError(f"Please select [time index] (integer) from {time_array}")
        self._time = value

    def _run_method(self, method: str, *args, **kwargs):
        return getattr(self, method)(*args, **kwargs)

    def nc_path(self, gem="ecogem", dim="2d"):
        "find netcdf file model_path, default is ecosystem model output"
        model_path = self.model_path
        nc_file = f"fields_{gem}_{dim}.nc"
        nc_path = join(model_path, gem, nc_file)
        if file_exists(nc_path):
            return nc_path

    def GENIE_lon(self, *args, **kwargs):
        return GENIE_lon(*args, **kwargs)

    def GENIE_lat(self, *args, **kwargs):
        return GENIE_lat(*args, **kwargs)

    def normal_lon(self, *args, **kwargs):
        return normal_lon(*args, **kwargs)

    def open_nc(self, path):
        "Note the time default is -1, and will depend on the instance, e.g., ForamVariable"
        return xr.open_dataset(path).isel(time=self.time)

    def auto_find_path(self, var):
        for gem in ["biogem", "ecogem"]:
            for dim in ["2d", "3d"]:
                nc_path = self.nc_path(gem, dim)
                if self.has_var(var, nc_path):
                    return nc_path
        raise ValueError("Variable not found, please check the spelling!")

    def has_var(self, var, nc_path):
        """
        check if variable exists
        """
        t = Dataset(nc_path, "r")
        if_exist = var in t.variables.keys()
        t.close()
        return if_exist

    def get_vars(self, var_name=None, *args, **kwargs):
        "return specified or all available variables"
        if not var_name:
            t = Dataset(self.nc_path(*args, **kwargs), "r")
            tmp = t.variables.keys()
            t.close()
            return tmp
        else:
            return xr.open_dataset(self.nc_path(*args, **kwargs))[var_name]

    def select_foram(self, foram_name: str):
        "a optimised version of select_var()"
        return ForamVariable(foram_name=foram_name, model_path=self.model_path)

    def select_var(self, var: str):
        return GenieVariable(var=var, model_path=self.model_path)

    def grid_mask(self, source="ecogem", Arctic=True, Med=True):
        """
        cGENIE continent mask array (by setting zero values),
        either calculated from existing data or using biogem.grid_mask
        """
        if source == "ecogem":
            data = self.select_var("eco2D_xGamma_T").array
            grid_mask = xr.where(~np.isnan(data), 1, 0).values
        elif source == "biogem":
            grid_mask = self.select_var("grid_mask").array
        else:
            raise ValueError("source only accept ecogem or biogem")

        if Arctic:
            grid_mask[34:36, :] = 0
        if Med:
            grid_mask[27:30, 25:30] = 0

        return grid_mask

    def marine_area(self):
        "grid area array in km2"
        grid_mask = self.grid_mask()
        grid_area = GENIE_grid_area()
        mask_area = grid_area * grid_mask

        return mask_area

    def marine_volume(self):
        "grid volume array in km3"
        grid_mask = self.grid_mask()
        grid_volume = GENIE_grid_vol()
        mask_volume = grid_volume * grid_mask

        return mask_volume

    def skill_score(self, cost_function="m_score", table_styler=True, *args, **kwargs):
        "summarised model skill score compared to modern observations"

        foram_abbrev = list(foram_names().keys())
        foram_fullname = tuple(foram_names().values())
        df = {
            "Biomass": [
                ForamVariable(i, self.model_path).biomass_c()._run_method(method=cost_function, *args, **kwargs)
                for i in foram_abbrev
            ],
            "Carbon Export": [
                ForamVariable(i, self.model_path).export_c()._run_method(method=cost_function, *args, **kwargs)
                for i in foram_abbrev
            ],
            "Relative Abundance": [
                ForamVariable(i, self.model_path)
                .export_c()
                .proportion().
                _run_method(method=cost_function, *args, **kwargs)
                for i in foram_abbrev
            ],
        }

        df = DataFrame(df, index=foram_fullname)

        df["Column Total"] = df.sum(axis=1)
        df.loc["Row Total", :] = df.sum(axis=0)

        if table_styler:
            df = df.style.set_caption(
                f"{cost_function} across foraminifer groups and variables compared to modern observation"
            ).text_gradient(cmap="viridis", subset=(df.index[0:4], df.columns[0:3]))

        return df

    def summarise(
        self,
        stat="nanmean",
        diff=False,
        diff_method="percentage",
        table_styler=True,
        *args,
        **kwargs,
    ):
        """
        summarise basic statistics of foraminifer groups
        """

        foram_abbrev = list(foram_names().keys())
        foram_fullname = tuple(foram_names().values())

        dic = {
            "Biomass": [
                ForamVariable(i, self.model_path).biomass_c()._run_method(method=stat)
                for i in foram_abbrev
            ],
            "Carbon Export": [
                ForamVariable(i, self.model_path).export_c()._run_method(method=stat)
                for i in foram_abbrev
            ],
            "Relative Abundance": [
                ForamVariable(i, self.model_path).export_c().proportion()._run_method(method=stat)
                for i in foram_abbrev
            ],
        }
        df = DataFrame(dic, index=foram_fullname)

        df["Column Total"] = df.sum(axis=1)
        df.loc["Row Total", :] = df.sum(axis=0)

        if diff:
            obs = obs_stat_bytype(type=stat, *args, **kwargs)
            diff = df - obs

            if diff_method == "percentage":
                diff_perc = diff / obs
                return diff_perc
            elif diff_method == "absolute":
                return diff

        if table_styler:
            df = df.style.set_caption(
                f"{stat} across foraminifer groups"
            ).text_gradient(cmap="viridis")

        return df

    def check_completeness(self):
        pass

    def eco_pars(self):
        """
        return ecophysiological parameter table
        """
        path = f"{self.model_path}/ecogem/Plankton_params.txt"
        df = read_fwf(path)
        return df

    def foram_POC(self):
        "Estimate total foraminiferal organic carbon flux rate"

        foram_poc = GenieArray()

        for foram in ["bn", "bs", "sn", "ss"]:
            foram_poc += self.select_foram(foram).export_c()

        poc_total = ForamCarbonFlux(self.model_path, "ALL_FORAM")
        poc_total.array = foram_poc.array

        return poc_total

    def foram_biomass(self):
        "Estimate total foraminiferal biomass"
        foram_biomass = GenieArray()

        for foram in ["bn", "bs", "sn", "ss"]:
            foram_biomass += self.select_foram(foram).biomass_c()

        biomass_total = ForamBiomass(self.model_path, "ALL_FORAM")
        biomass_total.array = foram_biomass.array

        return biomass_total

    def foram_calcite(self):
        "Estimate total foraminiferal inorganic carbon flux rate"
        return self.foram_POC().to_calcite()

    def foram_ldg(self, legend=True):
        "plot ldg: latitudinal diversity gradient"
        ## TODO: change colormap; change line width; add minor ticks
        ## add background grids
        foram_array_list = []
        foram_abbrev = list(foram_names().keys())
        foram_fullname = list(foram_names().values())

        for foram in foram_abbrev:
            array_1d = self.select_foram(foram).export_c().proportion().mean(axis=1)
            foram_array_list.append(array_1d)

        lat = GENIE_lat()

        fig = plt.figure(figsize=(6, 4))
        ax = fig.add_subplot(111)

        for n in range(4):
            ax.plot(lat, foram_array_list[n], label=foram_fullname[n])

        if legend:
            ax.legend(
                loc="lower center",
                ncol=2,
                bbox_to_anchor=(0.5, -0.35),
                edgecolor="black",
            )

        ax.set_xlabel("Latitude")
        ax.set_ylabel("Relative abundance")
        ax.tick_params(axis="y", direction="in")
        ax.tick_params(axis="x", direction="in")

        return ax

    def plot_currents(self, z_level=0):
        """
        plot velocity field of horizontal currents, default is the toppest layer
        Not able to add upon other field map yet.
        """
        fig = plt.figure()
        ax = fig.add_subplot(111, projection=ccrs.PlateCarree())

        v = self.select_var("phys_v").reassign_array().isel(zt=z_level)
        u = self.select_var("phys_u").reassign_array().isel(zt=z_level)
        m = np.hypot(u, v)

        lat = v.lat
        lon = v.lon
        lon2d, lat2d = np.meshgrid(lon, lat)

        ax.set_global()
        ax.stock_img()
        ax.coastlines()
        ax.quiver(lon2d, lat2d, m, u, v, transform=ccrs.PlateCarree())

        return ax

    def plot_biomass(self, *args, **kwargs):
        """
        quick wrapper function to plot carbon biomass for 4 foram groups
        """

        fig, axes = plt.subplots(
            nrows=2,
            ncols=2,
            figsize=(10, 7),
            subplot_kw=dict(projection=ccrs.EckertIV()),
        )

        varlst = ["eco2D_Plankton_C_01" + str(i) for i in range(6, 10)]
        most_max = max([self.select_var(varlst[i]).max() for i in range(4)])
        foram_fullnames = list(foram_names().values())

        for i, ax in enumerate(axes.flat):
            vdata = self.select_var(varlst[i]).array
            mean = np.nanmean(vdata)
            p = plot_genie(ax=ax, data=vdata, vmin=0, vmax=most_max, *args, **kwargs)
            ax.set_title(
                f"({string.ascii_lowercase[i]}) {foram_fullnames[i]} {mean:.2E}", pad=10
            )

        cbar = fig.colorbar(p, ax=axes, orientation="horizontal", pad=0.05, shrink=0.7)
        cbar.minorticks_on()
        cbar.set_label(r"carbon biomass $mmol C m$^{-3}$)", size=12)

        return p

    def plot_export(self):
        """
        quick wrapper function to plot carbon export for 4 foram groups
        """

        fig, axes = plt.subplots(
            nrows=2,
            ncols=2,
            figsize=(10, 7),
            subplot_kw=dict(projection=ccrs.EckertIV()),
        )

        varlst = ["eco2D_Export_C_01" + str(i) for i in range(6, 10)]
        most_max = max([self.select_var(varlst[i]).max() for i in range(4)])
        foram_fullnames = list(foram_names().values())

        for i, ax in enumerate(axes.flat):
            vdata = self.select_var(varlst[i]).array
            mean = np.nanmean(vdata)
            p = plot_genie(ax=ax, data=vdata, vmin=0, vmax=most_max)
            ax.set_title(
                f"({string.ascii_lowercase[i]}) {foram_fullnames[i]} {mean:.2E}", pad=10
            )

        cbar = fig.colorbar(p, ax=axes, orientation="horizontal", pad=0.05, shrink=0.7)
        cbar.minorticks_on()
        cbar.set_label(r"POC export at 80.8 m (mmol C m$^{-3}$ d$^{-1})$", size=12)

        return p

    def plot_abundance(self):
        """
        quick wrapper function to plot relative abundance for 4 foram groups
        """

        fig, axes = plt.subplots(
            nrows=2,
            ncols=2,
            figsize=(10, 7),
            subplot_kw=dict(projection=ccrs.EckertIV()),
        )

        varlst = list(foram_names().keys())
        foram_fullnames = list(foram_names().values())

        for i, ax in enumerate(axes.flat):
            vdata = self.select_foram(varlst[i]).export_c().proportion().array
            mean = np.nanmean(vdata) * 100
            p = plot_genie(ax=ax, data=vdata, vmin=0, vmax=1)
            ax.set_title(
                f"({string.ascii_lowercase[i]}) {foram_fullnames[i]} {mean:.2f}%",
                pad=10,
            )

        cbar = fig.colorbar(p, ax=axes, orientation="horizontal", pad=0.05, shrink=0.7)
        cbar.minorticks_on()
        cbar.set_label("Relative abundance", size=12)

        return p

    def barplot_comparison(self, *args, **kwargs) -> plt.axes:

        """
        Overview barplot of biomass and POC export compared to observed data
        :returns: matplotlib axes object
        """

        fname = list(foram_names().values())
        foram_abbrev = foram_names().keys()

        model_biomass_mean = [
            self.select_foram(i).biomass_c().nanmean() for i in foram_abbrev
        ]
        model_export_mean = [
            self.select_foram(i).export_c().nanmean() for i in foram_abbrev
        ]
        model_biomass_se = [self.select_foram(i).biomass_c().se() for i in foram_abbrev]
        model_export_se = [self.select_foram(i).export_c().se() for i in foram_abbrev]
        model_export_sum = [
            self.select_foram(i).export_c().sum().magnitude for i in foram_abbrev
        ]
        model_biomass_sum = [
            self.select_foram(i).biomass_c().sum().magnitude for i in foram_abbrev
        ]
        obs_biomass_mean = obs_stat_bysource("tow", *args, **kwargs).loc[:, "mean"]
        obs_biomass_se = obs_stat_bysource("tow", *args, **kwargs).loc[:, "se"]
        obs_export_mean = obs_stat_bysource("trap", *args, **kwargs).loc[:, "mean"]
        obs_export_se = obs_stat_bysource("trap", *args, **kwargs).loc[:, "se"]

        model_biomass_sum = DataFrame({"group": fname, "value": model_biomass_sum})
        model_export_sum = DataFrame({"group": fname, "value": model_export_sum})

        data_to_plot = [
            [
                [
                    model_biomass_mean,
                    model_biomass_se,
                    obs_biomass_mean,
                    obs_biomass_se,
                ],
                [model_export_mean, model_export_se, obs_export_mean, obs_export_se],
            ],
            [model_biomass_sum, model_export_sum],
        ]
        ## Plot starts
        fig, axes = plt.subplots(2, 2, figsize=(8, 6), sharex=True)
        bar_width = 0.3

        # The x position of bars
        x = np.arange(4)
        xlabels = [w.replace(" ", "\n") for w in fname]

        for i in range(2):
            for j in range(2):
                axes[i, j].yaxis.set_minor_locator(AutoMinorLocator(4))
                axes[i, j].set_axisbelow(True)
                axes[i, j].yaxis.grid(color="gray", linestyle="dashed")
                if i == 0:
                    axes[i, j].bar(
                        x - bar_width / 2,
                        data_to_plot[i][j][0],
                        width=bar_width,
                        color=sns.color_palette("Set1")[0],
                        edgecolor="black",
                        yerr=data_to_plot[i][j][1],
                        capsize=7,
                        label="model",
                    )

                    axes[i, j].bar(
                        x + bar_width / 2,
                        data_to_plot[i][j][2],
                        width=bar_width,
                        color=sns.color_palette("Set1")[1],
                        edgecolor="black",
                        yerr=data_to_plot[i][j][3],
                        capsize=7,
                        label="obs",
                    )

                    axes[i, j].set_xticks(x)
                    axes[i, j].set_xticklabels(xlabels, rotation=45, ha="right")
                    axes[i, j].legend()
                else:
                    sns.barplot(
                        data=data_to_plot[i][j],
                        x="group",
                        y="value",
                        ax=axes[i, j],
                        edgecolor="black",
                        palette="deep",
                    )
                    axes[i, j].set_xticklabels(xlabels, rotation=45, ha="right")
                    axes[i, j].set_xlabel("")
                    axes[i, j].set_xticks(x)
                    set_sns_barwidth(axes[i, j], bar_width)

        axes[0, 0].set_ylabel(r"mmol C m$^{-3}$")
        axes[0, 1].set_ylabel(r"mmol C m$^{-2}$ d$^{-1}$")

        axes[0, 0].set_title("(a)    global biomass mean/se", loc="left")
        axes[0, 1].set_title("(b)    global POC export mean/se", loc="left")
        axes[1, 0].set_title("(c)    globally integrated biomass", loc="left")
        axes[1, 1].set_title("(d)    globally integrated EP", loc="left")

        axes[1, 0].set_ylabel("Gt C")
        axes[1, 1].set_ylabel(r"Gt C yr$^{-1}$")

        fig.tight_layout()

        return axes

    def diff(self, model2compare, var):

        if isinstance(model2compare, GenieModel):
            B = model2compare
        else:
            B = GenieModel(model2compare)

        diff = self.select_var(var) - B.select_var(var)
        return diff

    def div(self, model2compare, var):
        B = GenieModel(model2compare)
        diff = self.select_var(var) / B.select_var(var)
        return diff

    def _ptf_presence(self, x, tol=1e-8):
        """
        to determine whethere a functional group present or not
        :param tol: threshold of biomass (mmol C/m3)
        """
        if np.isnan(x):
            return x
        elif x >= tol:
            return 1
        else:
            return 0

    def pft_count(self):
        "the number of plankton functional type"
        full_lst = list(self.get_vars())
        name_lst = [x for x in full_lst if "eco2D_Export_C" in x]
        n = len(name_lst)
        return n

    def pft_richness(self):
        """
        plankton functional group richness, note it is different from species richness,
        because there are more species in low size classes (i.e., body size-species richness relationship)
        """
        vfunc = np.vectorize(self._ptf_presence)
        total_sp = np.zeros((36, 36))
        n = self.pft_count()

        for i in range(n):
            name = f"eco2D_Plankton_C_0{i+1:02d}"
            # select
            arr = self.select_var(name).pure_array()
            # conditional mask
            sp = vfunc(arr)
            # sum
            total_sp += sp

        x = GenieArray()
        x.array = total_sp
        return x

    def select_pft(self, start=None, end=None, var="biomass", element="C"):

        # if not specify the groups, select all
        if not start and not end:
            start = 1
            end = self.pft_count()

        # biomass or export production
        if var == "biomass":
            v = "Plankton"
        elif var == "export":
            v = "Export"
        else:
            raise ValueError("Not correct variable")

        # loop and sum
        total = np.zeros((36, 36))
        end += 1
        for i in range(start, end):
            name = f"eco2D_{v}_{element}_0{i:02d}"
            arr = self.select_var(name).array
            total += arr

        # return
        x = GenieArray()
        x.array = total
        return x

    def cwm(self, trait: str, abundance="relative"):
        """community weighted mean, also representing position of community centroid (i,j),
        usually weighted in relative abundance
        """

        fg_dict = {"symbiont": ["sn", "ss"], "spine": ["bs", "ss"]}
        trait_dict = {"symbiont": [0.5, 1], "spine": [1, 1]}
        fg = fg_dict[trait]
        tr = trait_dict[trait]

        if abundance == "absolute":
            weighted_trait = (
                self.select_foram(fg[0]).biomass_c() * tr[0]
                + self.select_foram(fg[1]).biomass_c() * tr[1]
            )
            sum_abundance = (
                self.select_foram(fg[0]).biomass_c()
                + self.select_foram(fg[1]).biomass_c()
            )
            cwm = weighted_trait / sum_abundance
        elif abundance == "relative":
            cwm = (
                self.select_foram(fg[0]).export_c().proportion() * tr[0]
                + self.select_foram(fg[1]).export_c().proportion() * tr[1]
            )

        return cwm

    def fspec(self, *args, **kwargs):
        "Functional specialisation, distance from the origin (non-symbiont non-spinose)"
        fspec = distance(
            self.cwm(trait="spine", *args, **kwargs),
            self.cwm(trait="symbiont", *args, **kwargs),
        )

        x = GenieArray()
        x.array = fspec
        return x

    def fored(self):
        "functional over redundancy (only used for modern), unit: redundant species"
        fred = (
            self.select_foram("bn").export_c().proportion() * 5.25 / 47
            + self.select_foram("ss").export_c().proportion() * 11.25 / 47
        )

        return fred

    # def fdis(self, *args, **kwargs):
    # "functional dispersion"
    # weight (relative abundance/biomass)
    # weight = self.select_foram("bn").biomass_c()
    # each group's distance to centroid
    # dist =
    # weighted sum


class GenieVariable(GenieModel, GenieArray):
    def __init__(self, var, model_path):
        # initialise super class to inherite attributes
        self._var = var
        GenieModel.__init__(self, model_path=model_path)
        GenieArray.__init__(self)

    def _set_array(self):
        nc_path = self.auto_find_path(self._var)
        source_data = self.open_nc(nc_path)
        return source_data[self._var]


class ForamVariable(GenieModel):
    def __init__(self, foram_name, model_path):
        self.foram_name = foram_name
        GenieModel.__init__(self, model_path=model_path)

    def biomass_c(self):
        return ForamBiomass(model_path=self.model_path, foram_name=self.foram_name)

    def export_c(self):
        return ForamCarbonFlux(model_path=self.model_path, foram_name=self.foram_name)

    def sum_mscore(self):
        "compared to modern observations"
        return (
            self.export_c().m_score()
            + self.biomass_c().m_score()
            + self.export_c().proportion().m_score()
        )


class ForamBiogeochem(GenieModel, GenieArray):
    def __init__(self, model_path: str, foram_name: str, biogeo_var: str):
        self.biogeo_var = biogeo_var
        self.foram_name = foram_name
        GenieModel.__init__(self, model_path)
        self.ecogem_path = self.nc_path("ecogem", "2d")
        GenieArray.__init__(self)

    def _set_array(self):
        source_data = self.open_nc(self.ecogem_path)
        foramdict = foram_dict()
        for key, value in foramdict.items():
            if value[1] == self.foram_name and value[0] == self.biogeo_var:
                return source_data[key]

    def proportion(self):
        return ForamProportion(self.model_path, self.foram_name, self.biogeo_var)

    def m_score(self, *args, **kwargs):
        return quick_mscore(
            self.pure_array(), self.biogeo_var, self.foram_name, *args, **kwargs
        )

    def rmse(self, *args, **kwargs):
        return quick_rmse(
            self.pure_array(), self.biogeo_var, self.foram_name, *args, **kwargs
        )

    def cos_sim(self, *args, **kwargs):
        return quick_cos_sim(
            self.pure_array(), self.biogeo_var, self.foram_name, *args, **kwargs
        )

    def corr(self, *args, **kwargs):
        return quick_corr(
            self.pure_array(), self.biogeo_var, self.foram_name, *args, **kwargs
        )


class ForamBiomass(ForamBiogeochem):
    def __init__(self, model_path, foram_name):
        self.biogeo_var = "tow"
        super(ForamBiomass, self).__init__(model_path, foram_name, self.biogeo_var)
        self.unit = "mmol m$^-3$"

    def sum(self):
        C_ = molecular_weight("C")
        c = self.uarray().to_base_units()
        v = self.marine_volume().to_base_units()
        s = c * v
        s = s.to("mol").to("g", "chemistry", mw=C_ * ureg("g/mole")).to("Gt")

        return np.nansum(s)


class ForamCarbonFlux(ForamBiogeochem):
    def __init__(self, model_path, foram_name):
        self.biogeo_var = "trap"
        super(ForamCarbonFlux, self).__init__(model_path, foram_name, self.biogeo_var)
        self.unit = "mmol m$^-2$ d$^-1$"

    def _set_array(self):
        source_data = self.open_nc(self.ecogem_path)
        foramdict = foram_dict()
        for key, value in foramdict.items():
            if value[1] == self.foram_name and value[0] == self.biogeo_var:
                return source_data[key] * 80.8

    def to_calcite(self):
        """
        convert POC to Calcite (in mmol m-2 d-1) given POC:PIC:CaCO3 mol ratio = 100:36:36 (mass ratio = 100:36:300)
        """

        calcite = ForamCalcite(model_path=self.model_path)
        calcite.array = self.array * 0.36

        return calcite

    @ureg.with_context("bgc")
    def sum(self):
        # concentration data
        c = self.uarray().to_base_units()
        # make volume in pint type
        v = self.marine_area().to_base_units()
        # globall integrated value
        s = c * v

        # unit conversion
        C_ = molecular_weight("C")
        s = s.to("mol d^-1").to("g d^-1", "bgc", mw=C_ * ureg("g/mol")).to("Gt yr^-1")

        return np.nansum(s)


class ForamProportion(ForamBiogeochem):
    def __init__(self, model_path, foram_name, biogeo_var):
        self.biogeo_var = biogeo_var
        ForamBiogeochem.__init__(self, model_path, foram_name, biogeo_var)

    def _set_array(self):
        source_data = self.open_nc(self.ecogem_path)
        foramdict = foram_dict()
        all_foram_vars = []

        for key, value in foramdict.items():
            if value[0] == self.biogeo_var:
                all_foram_vars.append(source_data[key])

        total_foram = reduce(np.add, all_foram_vars)

        # ignore divided by 0
        # and set grid with total_foram == 0 to 0 instead of NA
        with np.errstate(divide="ignore", invalid="ignore"):
            one_foram = super()._set_array()
            proportion = np.divide(
                one_foram,
                total_foram,
                out=np.zeros_like(one_foram),
                where=total_foram != 0,
            )

        return proportion

    def m_score(self, observation="core", *args, **kwargs):
        return quick_mscore(
            self.pure_array(), observation, self.foram_name, *args, **kwargs
        )

    def rmse(self, observation="core", *args, **kwargs):
        return quick_rmse(
            self.pure_array(), observation, self.foram_name, *args, **kwargs
        )

    def cos_sim(self, observation="core", *args, **kwargs):
        return quick_cos_sim(
            self.pure_array(), observation, self.foram_name, *args, **kwargs
        )

    def corr(self, observation="core", *args, **kwargs):
        return quick_corr(
            self.pure_array(), observation, self.foram_name, *args, **kwargs
        )


class ForamCalcite(GenieArray, GenieModel):
    def __init__(self, model_path):
        "default empty array, need to assign the array manually, such as ForamCarbonFlux.to_PIC()"
        GenieModel.__init__(self, model_path=model_path)
        self.unit = "mmol m$^-2$ d$^-1$"

    @ureg.with_context("bgc")
    def sum(self):
        CaCO3 = molecular_weight("CaCO3")
        c = self.uarray().to_base_units()
        v = self.marine_area().to_base_units()
        s = c * v
        s = (
            s.to("mol d^-1")
            .to("g d^-1", "bgc", mw=CaCO3 * ureg("g/mol"))
            .to("Gt yr^-1")
        )

        return np.nansum(s)



import numpy as np
from scipy.spatial import distance
from netCDF4 import Dataset
from .data import obs_data
from .grid import mask_Arctic_Med


def safe_unveil(model_data):
    "get pure array from a numpy masked array object"
    if model_data.__class__ != np.ma.core.MaskedArray:
        return model_data
    else:
        return model_data.filled(np.nan)


def intersect_index(array1, array2, verbose=False):
    """
    Return the index where corresponding values are not
    nan in both input arrays. One then can filter the array
    by the output boolean array.
    """

    # If both are not NaN, return it
    array1 = safe_unveil(array1)
    array2 = safe_unveil(array2)

    indx_array = np.logical_and(~np.isnan(array1), ~np.isnan(array2))

    if verbose is True:
        num = indx_array.flatten()[indx_array.flatten() == True].shape[0]
        print("Summary: {} elements simultaneously exist.".format(num))

    return indx_array


def cal_mscore(data1, data2):
    """
    Calculate skill metric M-score. See more in the paper Watterson, I. G. (1996)

    Use 2D array as input, order causes no difference.
    NA will be removed.

    :param data1: data to compare, order does not count
    :param data2: data to compare, order does not count

    :type data1: numpy array
    :type data2: numpy array

    :returns: A float number
    """
    # get common set
    indx = intersect_index(data1, data2)
    sub_data1 = data1[indx]
    sub_data2 = data2[indx]

    # calculate M-score
    mse = np.square(np.subtract(sub_data1, sub_data2)).mean()
    v1 = sub_data1.var()
    v2 = sub_data2.var()
    g1 = sub_data1.mean()
    g2 = sub_data2.mean()

    mscore = (2 / np.pi) * np.arcsin(1 - mse / (v1 + v2 + np.square(g1 - g2)))

    return mscore


def cal_corr(data1, data2):
    """
    calculate pearson correlation coefficient for 2D array
    """
    indx = intersect_index(data1, data2)
    sub_data1 = data1[indx].ravel()
    sub_data2 = data2[indx].ravel()

    corr = pearson_r(sub_data1, sub_data2)
    return corr


def cal_cosine_similarity(data1, data2):
    """
    Calculate metric cosine similarity of two input arrays.

    Use 2D array as input, order causes no difference.
    """
    indx = intersect_index(data1, data2)
    sub_data1 = data1[indx]
    sub_data2 = data2[indx]

    if sub_data1.mean() != 0 and sub_data2.mean() != 0:
        cos_sim = 1 - distance.cosine(sub_data1.flatten(), sub_data2.flatten())
    else:
        cos_sim = np.nan

    return cos_sim


def cal_rmse(data1, data2):
    """
    Calculate Root Mean Sqaure Error (rmse, or rmsd) between two input arrays.

    Use 2D array as input, order causes no difference.
    """

    error_2d = data1 - data2
    error_1d = error_2d.ravel()[~np.isnan(error_2d.ravel())]
    rmse = np.sqrt(np.square(error_1d).mean())

    return rmse


def cal_crmse(data1, data2):
    """
    Calculate centred Root Mean Sqaure Error (rmse, or rmsd) between two input arrays. See Talor, K. E. (2001) JGR

    Use 2D array as input, order causes no difference.
    """
    indx = intersect_index(data1, data2)
    sub_data1 = data1[indx].ravel()
    sub_data2 = data2[indx].ravel()

    sigma1 = np.std(sub_data1)
    sigma2 = np.std(sub_data2)
    corr = pearson_r(sub_data1, sub_data2)
    crmse = sigma1**2 + sigma2**2 - 2 * sigma1 * sigma2 * corr

    return crmse


def get_foram_prop(file_path, var):
    """
    Quick calculation of [modelled] relative abundance, based on carbon export flux
    because of little difference between biomass and export.

    :param file_path: an netcdf file with all foram-related varialbes
    :param var: foram group abbrev: bn, bs, sn, ss

    :returns: a scalar value
    """

    f = Dataset(file_path)

    bn = safe_unveil(f.variables["eco2D_Export_C_016"][-1, :, :])
    bs = safe_unveil(f.variables["eco2D_Export_C_017"][-1, :, :])
    sn = safe_unveil(f.variables["eco2D_Export_C_018"][-1, :, :])
    ss = safe_unveil(f.variables["eco2D_Export_C_019"][-1, :, :])

    total_foram = bn + bs + sn + ss

    # ignore divided by 0
    with np.errstate(divide="ignore", invalid="ignore"):
        one_foram = locals()[var]
        proportion = np.divide(
            one_foram, total_foram, out=np.zeros_like(one_foram), where=total_foram != 0
        )

    f.close()

    return proportion


def quick_rmse(model, obs_source, var, *args, **kwargs):
    "A wrapper function to calculate RMSE"

    if hasattr(model, "values"):
        model = model.values

    masked_model = mask_Arctic_Med(model, policy="na")
    masked_data = mask_Arctic_Med(
        obs_data(obs_source, var, *args, **kwargs), policy="na"
    )

    return cal_rmse(masked_model, masked_data)


def quick_mscore(model, obs_source, var, *args, **kwargs):
    "A wrapper function to calculate M-Score"

    if hasattr(model, "values"):
        model = model.values

    masked_model = mask_Arctic_Med(model, policy="na")
    masked_data = mask_Arctic_Med(
        obs_data(obs_source, var, *args, **kwargs), policy="na"
    )

    return cal_mscore(masked_model, masked_data)


def quick_cos_sim(model, obs_source, var, *args, **kwargs):
    "A wrapper function to calculate cosine cimilarity"
    if hasattr(model, "values"):
        model = model.values

    masked_model = mask_Arctic_Med(model, policy="na")
    masked_data = mask_Arctic_Med(
        obs_data(obs_source, var, *args, **kwargs), policy="na"
    )

    return cal_cosine_similarity(masked_model, masked_data)


def quick_corr(model, obs_source, var, *args, **kwargs):
    "A wrapper function to calculate Pearson's correlation coefficient"
    if hasattr(model, "values"):
        model = model.values
        masked_model = mask_Arctic_Med(model, policy="na")
        masked_data = mask_Arctic_Med(
            obs_data(obs_source, var, *args, **kwargs), policy="na"
        )

    return cal_corr(masked_model, masked_data)


def pearson_r(data1, data2):
    """
    compute pearson correlation coefficient, which reflect linear similarity between two 1D arrays

    :param data1: 1D array
    :param data2: 1D array
    """

    # Compute corr matrix
    corr_mat = np.corrcoef(data1, data2)
    return corr_mat[0, 1]



from typing import Union
import pathlib

from xarray import open_dataset
import numpy as np
import pandas as pd
from scipy.stats import sem

from .grid import reassign_obs
from .utils import remove_outliers


def efficient_log(data: Union[int, float]) -> float:
    "keep NA, remove zeros"
    return np.where(data == 0, -10, np.log10(data))


def obs_data(source, var, stat=None, outlier_level=None):
    """
    Quickly fetch observational data from ForamData/data directory.
    The longitude coordinate will be assigned to fit GENIE output

    :returns: numpy 2D array
    :raise ValueError:
    """
    if source == "core":
        file_path = pathlib.Path(__file__).parent.parent / "data/ForCenS_regridded.nc"
    elif source == "tow":
        file_path = (
            pathlib.Path(__file__).parent.parent / "data/plankton_tow_regridded_abs.nc"
        )
    elif source == "trap":
        file_path = (
            pathlib.Path(__file__).parent.parent
            / "data/sediment_trap_regridded_organic.nc"
        )
    elif source == "MARGO":
        file_path = pathlib.Path(__file__).parent.parent / "data/MARGO_regridded.nc"
    else:
        raise ValueError("type must be core, obs or trap!")

    ds = open_dataset(file_path)
    long_name = foram_names()[var]
    obs = ds[long_name]
    modified_obs = reassign_obs(obs).to_numpy()

    if not outlier_level:
        pass
    else:
        modified_obs = remove_outliers(modified_obs, m=outlier_level)

    if not stat:
        return modified_obs
    else:
        mean = np.nanmean(modified_obs)
        sd = np.nanstd(modified_obs)
        se = sem(modified_obs, nan_policy="omit", axis=None)
        return mean, sd, se


def obs_stat_bysource(source, *args, **kwargs) -> pd.DataFrame:
    obs = []
    foram_abbrev = foram_names().keys()
    foram_fullnames = foram_names().values()

    for i in foram_abbrev:
        tmp = obs_data(source=source, var=i, stat="Yes", *args, **kwargs)
        obs.append(tmp)

    table = pd.DataFrame(obs, index=foram_fullnames, columns=["mean", "sd", "se"])
    return table


def obs_stat_bytype(type, *args, **kwargs) -> pd.DataFrame:
    tow = obs_stat_bysource("tow", *args, **kwargs).loc[:, type]
    trap = obs_stat_bysource("trap", *args, **kwargs).loc[:, type]
    core = obs_stat_bysource("core", *args, **kwargs).loc[:, type]
    # combination
    data = pd.concat([tow, trap, core], axis=1)
    data.columns = [
        "Biomass(mmol C/m3)",
        "Carbon Export (mmol C/m3/d)",
        "Relative Abundance",
    ]

    return data


# def get_diff_xarray(filename, type, var):
#     sorted_model_data, target_obs = get_comparable_data(filename, type, var)
#     diff_xarray = sorted_model_data - target_obs
#
#     return diff_xarray


def foram_names():
    """
    get a dictionary with foram abbrev (keys) and complete name (values).

    :returns: dictionary
    """
    foram_names = {
        "bn": "symbiont-barren non-spinose",
        "bs": "symbiont-barren spinose",
        "sn": "symbiont-facultative non-spinose",
        "ss": "symbiont-obligate spinose",
    }

    return foram_names


def foram_dict():
    """
    get a dictionary with foram variable name (key) and corresponding data source name(value[0]), abbrev (value[1])

    :returns: dictionary
    """
    foram_dict = {
        "eco2D_Plankton_C_016": ["tow", "bn"],
        "eco2D_Plankton_C_017": ["tow", "bs"],
        "eco2D_Plankton_C_018": ["tow", "sn"],
        "eco2D_Plankton_C_019": ["tow", "ss"],
        "eco2D_Export_C_016": ["trap", "bn"],
        "eco2D_Export_C_017": ["trap", "bs"],
        "eco2D_Export_C_018": ["trap", "sn"],
        "eco2D_Export_C_019": ["trap", "ss"],
    }

    return foram_dict


def get_calcite_rate(file, time=-1):
    """
    Estimate total foraminiferal inorganic carbon flux rate (g/year)

    :param file: string, netcdf output file
    :returns: 2D array
    """
    data = open_dataset(file).isel(time=time)

    df_bn = data.eco2D_Export_C_016
    df_bs = data.eco2D_Export_C_017
    df_sn = data.eco2D_Export_C_018
    df_ss = data.eco2D_Export_C_019

    POC_total = df_bn + df_bs + df_sn + df_ss  # mmol C/m3/day
    ind_rate = POC_total * 12 * 1e3 / 0.845  # ind/m3/day
    calcite_rate = ind_rate * 0.845 * 3 * 1e-3 * 80.8  # mg/m2/day
    calcite_g_yr = calcite_rate * 365 * 1e-3

    return calcite_g_yr


def filter_foramdf(dataframe, foram_group=None):
    """
    select data based on foraminifer group, either pass foraminifer abbrev
    or explicitly pass boolean values to `symbiosi` & `spinose`

    :param dataframe: pandas dataframe
    :param foram_group: abbreviation name of foraminifer

    :returns: a subsetted dataframe
    """
    foram_bool = {
        "bn": ["No", "No"],
        "bs": ["No", "Yes"],
        "sn": ["Yes", "No"],
        "ss": ["Yes", "Yes"],
    }

    trait = foram_bool[foram_group]

    query_string = "Symbiosis == '{}' & Spinose == '{}'".format(trait[0], trait[1])
    return dataframe.query(query_string)



from . import ureg, Q_
import pathlib

import numpy as np
import pandas as pd
import xarray
import geopandas as gpd
from shapely.geometry import Point


def lon_n2g(x):
    """
    Change parts of observational latitude [100, 180] to GENIE longitude [-260, -180]
    Note it isn't axisymmetric!
    """
    # TBD: add value range checker

    if x > 100 and x < 180:
        return x - 360
    else:
        return x


def lon_g2n(x):
    """
    Change parts of observational latitude [100, 180] to GENIE longitude [-260, -180]
    CANNOT simply use +80, or -80! It isn't axisymmetric!
    """
    # TBD: add value range checker

    if x < -180:
        return x + 360
    else:
        return x


def reassign_obs(data: xarray.Dataset) -> xarray.Dataset:
    return data.assign_coords({"lon": list(map(lon_n2g, data.lon.values))}).sortby(
        "lon"
    )


def reassign_GENIE(data: xarray.Dataset) -> xarray.Dataset:
    ## if hasattr(data, "assign_coords"):
    return data.assign_coords({"lon": list(map(lon_g2n, data.lon.values))}).sortby(
        "lon"
    )


def mask_Arctic_Med(array, policy="na"):
    """
    mask Arctic and Meditterean Sea in cGENIE modern continent configuration
    """
    if policy == "na":
        array[34:36, :] = np.nan
        array[27:30, 25:30] = np.nan
    elif policy == "zero":
        array[34:36, :] = 0
        array[27:30, 25:30] = 0

    return array


def GENIE_grid_mask(
    base="worjh2", basin="ALL", basin_lvl="", mask_Arc_Med=False, invert=False
):
    """
    Get a modern GENIE 36x36 mask array from input data.
    The input array is flipped (left/right flip -> up/down flip) for easy recognition

    :continent: worjh2, worlg4, worbe2, GIteiiaa, GIteiiaa, p0055c
    :basin: Atlantic/Pacific/Indian/ALL/Tanzania
    :basin_lvl: N/S/ALL, ALL means Southern Ocean section included

    :returns: GENIE grid array where continent/ice cap is 0 and ocean is 1, default is 'worjh2'
    """

    file_path = (
        pathlib.Path(__file__).parent.parent
        / f"data/mask_{base}_{basin}{basin_lvl}.txt"
    )
    grid_mask_raw = np.loadtxt(file_path, dtype=int)
    grid_mask = np.flip(np.fliplr(grid_mask_raw))

    if mask_Arc_Med:
        grid_mask = mask_Arctic_Med(grid_mask, policy="zero")

    if invert:
        grid_mask = ~grid_mask + 2

    return grid_mask


def GENIE_lat(N=36, edge=False):
    """
    return cGENIE latitude in log-sine normally degree resolution,
    if edge is False, then return midpoint
    """
    if edge:
        lat_edge = np.rad2deg(np.arcsin(np.linspace(-1, 1, N + 1)))
        return lat_edge
    else:
        lat = np.rad2deg(np.arcsin(np.linspace(-0.97222222, 0.97222222, N)))
        return lat


def GENIE_lon(N=36, edge=False):
    """
    return cGENIE longitude in 10 degree resolution,
    if edge is False, then return midpoint
    """
    if edge:
        lon_edge = np.linspace(-260, 100, N + 1)
        return lon_edge
    else:
        lon = np.linspace(-255, 95, N)
        return lon


def GENIE_depth(edge=False):
    z_edge = np.array(
        [
            0.0000,
            80.8407,
            174.7519,
            283.8467,
            410.5801,
            557.8040,
            728.8313,
            927.5105,
            1158.3124,
            1426.4307,
            1737.8987,
            2099.7254,
            2520.0527,
            3008.3391,
            3575.5723,
            4234.45166,
            5000.0000,
        ]
    )
    if edge:
        return z_edge
    else:
        z = np.array([(z_edge[i] + z_edge[i + 1]) / 2 for i in range(len(z_edge) - 1)])
        return z


def normal_lon(N=36, edge=False):
    """
    Normal longitude in 10 degree resolution,
    if edge is False, then return midpoint
    """
    if edge:
        lon_edge = np.linspace(-180, 180, N + 1)
        return lon_edge
    else:
        lon = np.linspace(-175, 175, N)
        return lon


def GENIE_grid_area():
    """
    get grid area array with unit of km^2, not masked with continent/Arctic/Mediterranean!
    """
    # x level
    lon_len = np.repeat(10, 36)
    lon_len_km = lon_len * 111.320  # cos(rad(lat)) needed

    # y level
    lat_edge = np.rad2deg(np.arcsin(np.linspace(-1, 1, 37)))
    lat_width = [lat_edge[i + 1] - lat_edge[i] for i in range(36)]
    lat_width_km = np.array(lat_width) * 110.574

    # multiplication in km^2
    # np.outer: col x row -> out[i, j] = a[i] * b[j]
    grid_area = np.outer(lat_width_km, lon_len_km)  # check the order!
    # add the cos(rad(lat)) term
    lat_rad = np.arcsin(np.linspace(-0.97222222, 0.97222222, 36))
    for i in range(36):
        grid_area[i,] = (
            grid_area[
                i,
            ]
            * np.cos(lat_rad)[i]
        )

    unit = ureg("km^2")

    return grid_area * unit


def GENIE_grid_vol():
    """
    get grid_volume array with unit of km^3, not masked with continent!
    """

    # z_edge level in 1st layer
    vertical_width_km = Q_(80.8 / 1000, "km")

    # grid volumn: km^3
    grid_volume = GENIE_grid_area() * vertical_width_km

    return grid_volume


# check grid area (1)
# print(marine_area.sum()*1e-8) #around 3.7

# check grid area (2)
# np.nansum(np.where(mask_array == 1, grid_area, np.nan))*1e-8 #same

# check grid area (3) plot out
# plt.pcolormesh(marine_area)
# plt.colorbar()


# check grid area (4) compare to biogem grid area
# bgem_grid_area = xr.load_dataset("~/Downloads/fields_biogem_2d.nc")
# np.nansum(np.where(mask_array == 1, bgem_grid_area, np.nan))*1e-6*1e-8


# def interp_GENIE(file_path, var):
#     """
#     Interpolate GENIE into 360x360 grids using xesmf API. Depends on Xarray.
#     Known limitations:
#         (1) unexpected NAs grids in boundaries, e.g, Arctic and 95-degree longitude;
#         (2) thus necessary offset in the map;
#         (3) ugly way to draw continent lines

#     :param file_path: string, netcdf file
#     :parm var: string, variable name
#     """

#     # ------Prepare finer grids-------------
#     coarse_data = open_dataset(file_path)
#     # get 10 times higher resolution
#     #fine_lat = np.rad2deg(np.arcsin(np.linspace(-0.97222222, 0.97222222, 360)))  # sin(x) -> radian -> degree
#     #fine_lon = np.linspace(coarse_data.lon[0], coarse_data.lon[-1], 360)
#     fine_lat = GENIE_lat(360)
#     fine_lon = GENIE_lon(360)

#     # prepare finer grid
#     fine_data = Dataset(
#         {"lat": (["lat"], fine_lat),
#          "lon": (["lon"], fine_lon)}
#     )

#     # ----- Mask coarse & fine grids ---------
#     coarse_data['mask'] = np.where(~np.isnan(coarse_data[var].isel(time=0)), 1, 0)  # NA data to 0, otherwise 1

#     # mask output grid
#     mask_array_coarse = coarse_data['mask'].values
#     mask_array_fine = np.zeros((360, 360))

#     #359 or 360?
#     for i in range(360):
#         for j in range(360):
#             if mask_array_coarse[i // 10, j // 10] != 0:
#                 mask_array_fine[i, j] = 1

#     mask_array_fine_da = DataArray(
#         data=mask_array_fine,
#         dims=["lat", "lon"],
#     )

#     fine_data['mask'] = mask_array_fine_da

#     # --------- Regrid -----------
#     regridder = xe.Regridder(coarse_data, fine_data, method="bilinear", extrap_method="inverse_dist")
#     fine_data = regridder(coarse_data[var])

#     return fine_data


def regrid_lat(x):

    """
    Transform <latitude> into cGENIE resolution to facilitate comparison between
    model and observational data.
    """
    if x >= -90 and x <= 90:
        lat_edge = np.rad2deg(np.arcsin(np.linspace(-1, 1, 37)))
        lat = GENIE_lat()

        for i in range(36):
            if x > lat_edge[i] and x <= lat_edge[i + 1]:
                x = lat[i]
    else:
        raise ValueError("Latitude must be in [-90,90]")

    return x


def regrid_lon(x):

    """
    Transform <longitude> into cGENIE resolution to facilitate comparison between
    model and observational data.
    """

    if x >= -180 and x <= 180:
        lon_edge = np.linspace(-180, 180, 37)
        for i in range(36):
            if x > lon_edge[i] and x <= lon_edge[i + 1]:
                x = (lon_edge[i] + lon_edge[i + 1]) / 2  # middle value in the bin
    else:
        raise ValueError("Longitude must be in [-180,180]")

    return x


def new_basin_mask(mask_array, basin, filename):
    """
    a dirty way to copy the basin mask from worjh2 and facilitate further manual modification
    """

    x = mask_array + GENIE_grid_mask(base="worjh2", basin=basin)
    o = np.where(x > 1, 1, 0)
    o = np.flip(np.fliplr(o))
    np.savetxt(filename, o, fmt="%i")
    print(f"array saved in {filename}")


def regrid_dataframe(
    dataframe,
    low_threshold=None,
    new_low_bound=None,
    high_threshold=None,
    new_high_bound=None,
):

    """
    Regrid a dataframe within certain format to cGENIE grids

    Input: A dataframe with three necessary columns: "Latitude", "Longitude", "Observation".
    The other columns will be unselected. Note that observation column is the data you want to process.
    Rename dataframe by using command: `df = df.rename({"old name": "new name"}, axis='columns')`

    Output:
    A 36x36 2D array.

    Optional parameters:
    low_threshold/new_low_bound: change the data that is smaller than low_threshold to new_low_bound.
    For example, covert any data < 0.03 to 0. Work same to high_threshold/new_high_bound.
    """

    df = dataframe.copy()

    # subset
    df = df[["Longitude", "Latitude", "Observation"]]

    # drop NAN
    df = df.dropna(axis="rows", how="any")

    # regrid coordinate
    df.loc[:, "Latitude"] = df.loc[:, "Latitude"].apply(regrid_lat)
    df.loc[:, "Longitude"] = df.loc[:, "Longitude"].apply(regrid_lon)

    # group and aggregate (still in long format)
    df_agg = df.groupby(["Longitude", "Latitude"]).agg("mean")

    # Create a all-nan data frame with full cGENIE grids
    lat = GENIE_lat()
    lon = normal_lon()
    data = np.zeros([36 * 36])
    index = pd.MultiIndex.from_product([lon, lat], names=["Longitude", "Latitude"])
    df_genie = pd.DataFrame(data, index=index, columns=["Observation"])
    df_genie["Observation"] = df_genie["Observation"].replace(0, np.nan)

    # transform the tuple elements in multindex to lists
    df_genie_index_list = [list(item) for item in df_genie.index.values]
    df_agg_index_list = [list(item) for item in df_agg.index.values]

    # copy aggregated dataframe to full-grid dataframe, one by one
    # if the returned dataframe is full of NA, it's very likely from this step
    for i in df_genie_index_list:

        longitude = i[0]
        latitude = i[1]

        if i in df_agg_index_list:
            df_genie.loc[longitude, latitude] = df_agg.loc[longitude, latitude]
        else:
            df_genie.loc[longitude, latitude] = np.nan

    # filter data if necessary
    if (low_threshold is not None) and (new_low_bound is not None):
        df_genie.loc[
            df_genie.Observation < low_threshold, "Observation"
        ] = new_low_bound

    if (high_threshold is not None) and (new_high_bound is not None):
        df_genie.loc[
            df_genie.Observation > high_threshold, "Observation"
        ] = new_high_bound

    # long -> wide data format
    df_genie_wide = df_genie.pivot_table(
        values="Observation", index="Latitude", columns="Longitude", dropna=False
    )

    return df_genie_wide


def detect_ocean(lon, lat):
    "use point-in-polygon strategy to detect ocean basin according to lon/lat"

    file_path = pathlib.Path(__file__).parent.parent / "data/oceans/oceans.shp"

    oceans = gpd.read_file(file_path)
    p = Point(lon, lat)

    ocean_name = oceans[oceans.contains(p)].Oceans.values
    if ocean_name.size > 0:
        return ocean_name[0]
    else:
        return ""
